function [cn, z, info]=crawfnos(A, opts)
% CRAWFNOS computes the Crawford number of a square matrix.
%
%   cn = crawfnos(A/Afun) returns the Crawford number of a matrix A,
%               or a linear operator Afun(structure) with elements
%               Afun.matvec(x) = A*x    matrix-vector product
%               Afun.rmatvec(x) = A'*x  conjugate transpose matrix-vector 
%               Afun.matmat(V) = A*V    matrix-matrix product
%               Afun.n = size A
%
%
%
%   [cn, z, info] = crawfnos(A) also returns the point z, where the 
%       Crawford number is attained in the numerical range of A, 
%       i.e., cn = abs(z). 
%       The information structure
%       - info.its:     number of calls to eigenvalue solvers.
%       - info.lb (ub): lower (upper) bounds of cn.
%       - info.conv:    true if converged
%       - info.msg:     error message 
%       - info.THETA:   theta_k of each iteration
%       - info.PHI:     phi_k of each iteration
%       - info.LAMBDA:  lambda_k of each iteration
%
%
%
%   cn = crawfnos(A, opts) specifies options in the computation
%       - opts.tol:     tolerance |ub-lb| < tol*max([1, abs(cn)])
%                       (default by 1.0E-13)
%       - opts.dodisp:  display computation history if true 
%                       (default false)
%       - opts.mode:    computation mode
%                       'b' - basic subspace method (default)
%                       'r' - reduced subspace method
%                       'g' - computes the 'generalized' Crawford number.
%                               In this mode cn is negative if 0 is 
%                               contained in the numerical range of A.
%       - opts.theta0:  initial searching angle (default 0)
%       - opts.dobisect: do bisection safeguarding if true (default false)
%       - opts.redmode: Reduced problem solver.. default criss-cross
% 
%   References: 
%       [1] D. Kressner, D. Lu, and B. Vandereycken, Subspace acceleration for
%           the Crawford number and related eigenvalue optimization problems,
%           manuscript available at http://www.unige.ch/~dlu/.
%   
%   Authors:
%       Ding Lu and Bart Vandereycken, University of Geneva.
%
%   Modified:
%       April 25, 2017.


if nargin < 2, opts = []; end;
[tol, dodisp, mode, theta0, dobisect, redmode] = CheckInput(opts);
mode_in = 1*(mode=='f') + 2*(mode=='r') + 3*(mode=='g');

% ------------------------------------
%   Call computational routine
% ------------------------------------
[cn, z, info]=crawfnos_sg(A, tol, dodisp, mode_in, dobisect, theta0, redmode);


return
% END

% ===================================
%  CHECK INPUT 
% ===================================
function [tol, dodisp, mode, theta0, dobisect, redmode] = CheckInput(opts)
if ~isfield(opts, 'tol') || isempty(opts.tol)
    tol = 1.0E-13;
else
    tol = opts.tol;
end
if ~isfield(opts, 'dodisp') || isempty(opts.dodisp)
    dodisp = false;
else
    dodisp = opts.dodisp;
end
if ~isfield(opts, 'mode') || isempty(opts.mode)
    mode = 'b';
else
    mode = opts.mode;
end
if ~isfield(opts, 'theta0') || isempty(opts.theta0)
    theta0 = 0;
else
    theta0 = opts.theta0;
end
if ~isfield(opts, 'dobisect') || isempty(opts.dobisect)
    dobisect = false;
else
    dobisect = opts.dobisect;
end

if ~isfield(opts, 'redmode') || isempty(opts.redmode)
    redmode = 'c';
else
    redmode = opts.redmode;
end

return;

