function [cn, z, info]=crawfnos_sg(Afun, tol, dodisp, mode, dobisect, theta0, redsolver, mp_opts)
% Subspace Crawford number computation.
%
%   Input
%       Afun    matrix A, or linear operator (structure) with elements
%               Afun.matvec(x) = A*x    matrix-vector product
%               Afun.rmatvec(x) = A'*x  conjugate transpose matrix-vector 
%               Afun.matmat(V) = A*V    matrix-matrix product
%               Afun.n = size A
%
%       tol     tolerance |ub-lb| < tol*max([1, abs(cn)])
%       dodisp  display iteration information, if true
%       mode    1: basic subspace, 
%               2: reduced subspace,
%               3: Generalized Crawford number basic subspace.
%                   In this mode cn is negative if 0 is 
%                   contained in the numerical range of A.
%       dobisect    do bisection safeguard if true (not available in mode 3)
%       theta0  initial theta0
%       redsolver   solver for the reduced Crawford number
%                   'b': bisection solver
%                   'u': Uhlig's alg.
%                   'c': Criss-cross (default)
%       mp_opts     This input facilitates simulation by
%                   multiprecision computation. 
%                   opts.mpi =  mp('pi')
%                   opts.mpeps =  mp('eps')
%       
%   Output
%       cn      Crawford number
%       z       the point where the Crawford number is attained in the 
%               numerical range of A, i.e., cn = abs(z). 
%       info    information structure
%       - info.its:     number of calls to eigenvalue solvers.
%       - info.lb (ub):    lower (upper) bounds of cn.
%       - info.conv:    true if converged
%       - info.msg:     error message 
%       - info.THETA:   theta_k of each iteration
%       - info.PHI:     phi_k of each iteration
%       - info.LAMBDA:  lambda_k of each iteration
%
%
%   References: 
%       [1] D. Kressner, D. Lu, and B. Vandereycken, Subspace acceleration for
%           the Crawford number and related eigenvalue optimization problems,
%           manuscript available at http://www.unige.ch/~dlu/.
%   
%   Authors:
%       Ding Lu and Bart Vandereycken, University of Geneva.
%
%   Modified:
%       April 25, 2017.
%
%   Possible improvements:  1.  use direction vector instead of angle
%                           2.  stopping criteria take into account of eigenvalue
%                               accuracy.
%                           3.  Stagnation check for 3 vector version
%                   

if nargin == 8   % mp tool box setting
    mpi = mp_opts.mpi;
    tol_eps = mp_opts.mpeps;
else
    mpi = pi;
    tol_eps = eps;
end

if ~isstruct(Afun)
    maxit = length(Afun);
else
    maxit = Afun.n;
end

tol_rd = tol/100;   % tolerance for the reduced Crawford number
neigval = 0;        % number of eigenvalue evaluation
%theta1 = theta0 + mpi/2; %% Old version, use fixed theta1

% initialize base matrix 
[Q, v0, v1, theta1, l0, l1, phi0] = init_base(Afun, theta0, mpi);
neigval = neigval + 2;

% lower and upper bound of Crawford number
lb = max([l0, l1]); ub = phi0;

% Display results
if dodisp 
    fprintf(' ------\n');
    fprintf(' its %3i: cn = %6.5E, absolute error = %2.1E \n',... 
        1, phi0, ub - lb);
end

% computation history
ITV = [theta0, theta0 + 2*mpi];
THETA = [theta0; theta1];
PHI = [phi0];
LAMBDA = [l0; l1];

% input check
if mode == 2; V = [v0, v0]; v = v1; end
if mode == 3; dobisect = false; end

% --------------------- %
%   MAIN LOOP
% --------------------- %
for i = 1 : maxit 
    ub0 = ub;

	% Compute the reduced Crawford number
    if isstruct(Afun)
        Ak = Q' * Afun.matmat(Q);
    else
        Ak = Q' * (Afun*Q);
    end
    
    if redsolver == 'u' 
        % solve by Uhlig's alg
	    [cni, rai, ~] = crawfno_moduhlig(Ak, 0, 0, tol_rd, tol_rd, 10000, mpi);
        ub = min([cni, ub]);
        rai = rai * sign(cni); % in Uhlig's alg, closest point is cni*rai
        if ub == cni, ubdir = rai(1) + 1i*rai(2); end;
    elseif redsolver == 'b'
        % by bisection
	    [cni, rai_c, info_rd] = crawfno_bisect(Ak, tol_rd);
        rai = [real(rai_c); imag(rai_c)];
        ub = min([cni, ub]);
        if ub == cni, ubdir = rai_c; end;
    else 
        % criss cross
        [cni, rai_c] = numr(Ak, 0, 0, 0, mpi, tol_rd, max([lb, 0]));
        rai_c = rai_c/abs(rai_c);
        rai = [real(rai_c); imag(rai_c)];
        ub = min([cni, ub]);
        if ub == cni, ubdir = rai_c; end;
    end

    % Display results
    if dodisp 
        fprintf(' its %3i: cn = %6.5E, absolute error = %2.1E \n',... 
            neigval, cni, abs(ub - lb));
    end

    THETA = [THETA; angle(rai(2)*1i + rai(1))];
    PHI = [PHI; cni];

    % Check breakdown
    if mode~= 3 && cni <= 0 
        info.conv = true;
        info.msg = 'Crawford number == 0!';
        cn = 0; z = 0;
        break;
    end

    % Stopping criteria
    if i > 1 && ~isinf(ub) 
        if ub - lb  < tol*max([norm(Ak), abs(ub)])
            cn = ub; z = abs(cn)*ubdir;
            info.conv = true;
            info.msg = 'Crawford number accuracy tolerance achieved ';
            break; 
        elseif mode~=2 && ub0 - ub < tol_eps*max([abs(ub0),1])
            cn = ub; z = abs(cn)*ubdir; 
            info.conv = false;
            info.msg = 'Reduced Crawford number stagnating. ';
            break; 
        end
    end
    if i == maxit
        cn = ub; z = abs(cn) * ubdir; 
        info.conv = false;
        info.msg = 'Maximun iteration number reached!';
        break; 
    end

    % Interval update (for mode 1 and 2)
    if mode~= 3 && inrange(ITV(end,1), THETA(end-1), THETA(end), mpi)
        ITV = [ ITV; ITV(end,1), THETA(end-1)];
        if mode == 2, V = [V(:,1), v]; end
    elseif mode~=3
        ITV = [ ITV; THETA(end-1), ITV(end,2)];
        if mode == 2, V = [v, V(:,2)]; end
    else
        % mode 3: do not track the interval

    end

    % Bisection (optional)
    if dobisect && adist(ITV(end,:), mpi) > 0.5 * adist(ITV(end-1,:), mpi)
        THETA(end) = amid(ITV(end-1,:), mpi);
    end

	% Subspace update
	[v, E] = eval_phi(Afun, THETA(end));
    neigval = neigval + 1;
    LAMBDA = [LAMBDA; E];

    if mode~=2 
        Q = orth([Q,v]);
    else
        Q = orth([V,v]);
    end
    lb = max([E, lb]);
end

% Display results
if dodisp 
    if info.conv
        fprintf(' ------\n  Subspace algorithm converged in %2i its:\n\tCrawford number \t%16.15E\n  \tAbsolute error  \t%2.1E \n\tCalls to eig/eigs \t%i\n ------\n', ...  
        neigval, cn, abs(ub - lb), neigval);
    else
        fprintf([' ------\n  Subspace algorithm fails to converge: \n\t', info.msg, '\n ------']);
    end
end

info.its = neigval;
info.THETA = THETA;
info.PHI = PHI;
info.LAMBDA = LAMBDA;
info.lb = lb;
info.ub = ub;

return;

%==============================================================================
function [V1, E1] = eval_phi(Afun, theta)
% [V1, E1] = eval_phi(Afun, theta) evaluates the smallest eigenvalue E1 and
%   the corresponding eigenvector V1 of the Hermitian matrix
%   Hfun = ( Afun + Afun' ) * .5 * cos(theta) + ( Afun' - Afun) * .5i * sin(theta)

if isstruct(Afun)
    Hfun = @(x) 0.5*(cos(theta)-1i*sin(theta))*Afun.matvec(x) + 0.5*(cos(theta)+1i*sin(theta))*Afun.rmatvec(x);
    [V1, E1] = eigval(Hfun, Afun.n, 2);  % compute smallest eigenvalue, and eigenvector.
elseif length(Afun) > 800
    Hfun = @(x) 0.5*(cos(theta)-1i*sin(theta))*Afun*x + 0.5*(cos(theta)+1i*sin(theta))*Afun'*x;
    [V1, E1] = eigval(Hfun, length(Afun), 2);  % compute smallest eigenvalue, and eigenvector.
else
    H = 0.5*(cos(theta)-1i*sin(theta))*Afun + 0.5*(cos(theta)+1i*sin(theta))*Afun';
    [V1, E1] = eigval(H, length(Afun), 1);  % dense solver
end
return
% END OF EVAL_PHI

%==============================================================================
function [V, E] = eigval(Hfun, n, mode)
% [V, E] = eigval(H) evaluates smallest eigenvalues/vector of a Hermitian matrix H.
if mode == 1
    Hfun = 0.5*(Hfun+Hfun'); % symmetry
    [VV, EE] = eig(Hfun);
    EE = diag(EE);
    [~, idx] = min(EE);
    V = VV(:,idx(1));
    E = EE(idx(1));
else
    opts.p = 30;
    opts.maxit = 1600;
    opts.isreal = false;
    opts.issym = true;
    [V,E] = eigs(Hfun, n, 1,'sr',opts); 
end
return
% END OF EIGVAL

%==============================================================================
function [V1, v0, v1, theta1, e0, e1, cn0] = init_base(Afun, theta0, mpi)
% [V, v0, v1] = init_base(Afun, theta) evaluates the initial base matrix.
[v0, e0] = eval_phi(Afun, theta0);
if isstruct(Afun)
    pt = v0'*Afun.matvec(v0);
else
    pt = v0'*Afun*v0;
end
theta1 = angle(pt); theta1 = mod(theta1+2*mpi, 2*mpi);
[v1, e1] = eval_phi(Afun, theta1);
V1 = orth([v0,v1]);
cn0 = abs(pt);

return;
% END OF INIT_BASE

%==============================================================================
function info = inrange(l, u, p, mpi)
% info = inrange(l, u, p), return true if p is within l, u, in 2pi periodic 
%   sense.
ll = 0;
uu = mod(u-l+4*mpi, 2*mpi);
pp = mod(p-l+4*mpi, 2*mpi);
info = (ll < pp) * (pp < uu);
return;

%==============================================================================
function da = adist(Ang, mpi)
da = Ang(2) - Ang(1);
da = mod(da+2*mpi, 2*mpi);
return;

%==============================================================================
function ag = amid(Ang, mpi)
ag = Ang(1) + 0.5*adist(Ang, mpi);
return;
