function [cn, dir, info] = crawfno_bisect(A, tol)
% [cn, dir, info] = crawfno_bisect(A, tol)
%   computes the Crawford number of a matrix A by bisection method,
%   applied to the angle range [l, u] containing the optimum of vphi(theta).
%
%  On output:
%   cn      --  Crawford number (lower bound)
%               NOTE:   cn < 0 implies 0 is within the numerical range,
%                       the value of the number generally has no special meaning.
%   dir     --  Optimal points direction on the boundary of numerical range, s.t.,
%               tp = cn*dir
%   info    --  Upper bound
%
%  On input:
%   A       --  Data matrix
%   tol     --  tolerance (default by sqrt(eps))
%
%
%  Created 09-12-2016, by Ding.Lu@unige.ch
%
%  POSSIBLE IMPROVEMENT:    stopping criteria take into account of eigenvalue
%                           accuracy.

if nargin<2, tol = eps*2*pi; end;
tol = max([tol, eps*2*pi]);

maxit = 56;

% initial angle range, and touching points
l = -pi; u = pi; c = 0;
[pl, El] = eval_touching_point(A, l);
[pc, Ec, sc] = eval_touching_point(A, c);
pu = pl;

% keep track of lower and upper bound of Crawford number
if El > Ec
    lb = El; ltheta = l;
else
    lb = Ec; ltheta = c;
end
ub = inf;
utheta = inf;

% ----------------------- %
%       MAIN LOOP
% ----------------------- %

for i = 1:maxit

    cc = ClosePointToZero(pl, pc);  % optimal point on line pl--pc
    if abs(cc) < ub, ub = abs(cc); utheta = angle(cc); end % upper bound
    cr = ClosePointToZero(pu, pc);  % optimal point on line pu--pc
    if abs(cr) < ub, ub = abs(cr); utheta = angle(cr); end % upper bound

    % stopping criteria:    1. interval accuracy achieved, maxit reached;
    %                       2. Crawford number reach eps precision 
    if u - l < tol || i == maxit 
        %cn = ub; cn = ub; dir = exp(c*1i);
        cn = Ec; dir = exp(c*1i);
        info.conv = true;
        info.msg='Angle interval accuracy reached tol.';
        break
%    elseif ub - lb < eps*abs(ub)
%        cn = ub; cn = ub; dir = exp(c*1i);
%        info.conv = true;
%        info.msg='Crawford number accuracy reached eps.';
%        break
    end

    
    % Bisection : divide interval by c
    if Ec > 0 && i > 1
        % use eigenvalue derivative to bisect
        if sc > 0
            l = c; pl = pc;
        else
            u = c; pu = pc;
        end
    else 
        % use reduced function to bisect
        d0 = u - l;
        islc = (angle(cc) <= c) *  (angle(cc) >= l);    % is in [l, c] ?
        iscu = (angle(cr) <= u) * (angle(cr) >= c);     % is in [l, c] ?

        if (~islc && ~iscu) 
%            % Out of bracket? 
%            %   Run fminbnd, or return error message
%            opts0.TolX = eps;
%            cc = fminbnd(@(x) -vphi(A,x), l, u, opts0);
%            [lb, tp] = vphi(A, cc);
%            ltheta = cc;
%            if abs(tp) < ub, ub = abs(tp); utheta = angle(tp); end;
%            if ub - lb < eps*max([abs(ub),1])
%                cn = ub; dir = exp(cc*1i);
%                info.conv = true;
%                info.msg='Crawford number accuracy reached eps.';
%            else
%                cn = ub; dir = exp(cc*1i);
%                info.conv = false;
%                info.msg='Angle falls out of bracket, possibly due to numerical error.';
%            end;

            cn = Ec;
            dir = exp(c*1i);
            info.conv = false;
            info.msg='Angle falls out of bracket, possibly due to numerical error.';
            break
           
        elseif islc % in [c, u]
            u = c;
            pu = pc;
        else        % in [l, c]
            l = c;
            pl = pc;
        end
    end

    % Update midpoint, and new touching point
    c = 0.5*(l+u);
    [pc, Ec, sc] = eval_touching_point(A, c);
    if Ec>lb, lb = Ec; ltheta = c; end  % update upper bound of Crawford number
end

info.lb = lb;
info.ltheta = ltheta;
info.ub = ub;
info.utheta = utheta;
info.its = i;

return;
% END OF CRAWFNO_BISECT

%==============================================================================
function [e, p] = vphi(A, theta)
[p, e] = eval_touching_point(A, theta);
return

function [p, E1, ss] = eval_touching_point(A, theta);
% p = eval_touching_point(A, theta) evaluates the 'closest' touching point p
%   of the numerical range of A, in direction theta 
H = ( A + A' ) * .5 * cos(theta) + ( A' - A) * .5i * sin(theta); 
[V1, E1] = eigval(H);  % compute smallest and largest eigenvector.
if length(E1) == 1
    % single eig
    p = V1'*A*V1; % touching points, row vector
    ss = -real(p)* sin(theta) + imag(p) * cos(theta); 
else
    % possible clustered eig
    E1 = E1(1);
    P = V1'*A*V1; 
    p = P(1,1); 
    ss = -real(p)* sin(theta) + imag(p) * cos(theta); 
    P = diag(P); p = ClosePointToZero(P(1), P(2));
end


return
% END OF EVAL_TOUCHING_POINT

%==============================================================================
function [V, E] = eigval(A)
% E = eigval(A) evaluates smallest eigenvalues/vector of a Hermitian matrix A.
if length(A) < 800
    [VV, EE] = eig(A);
    EE = diag(EE);
    [EE, idx] = sort(EE, 'ascend');
    idxx = find(EE-EE(1) < abs(EE(end)) * 1.0E-8);
    V = VV(:,idx(idxx)); 
    E = EE(idx(idxx));
else
    opts.p = 30;
    opts.maxit = 1600;
    [V,E] = eigs(A,1,'sr',opts); 
end
return

% END OF EIGVAL

%==============================================================================
function cc = ClosePointToZero(pl, pc)
% p = ClosePointToZero(pl, pu) evaluates closest point on the segment pl--pu to
%   zero.
dd = pl-pc;
alpha = -real(pc'*dd)/real(dd'*dd);
if alpha>=0 && alpha<=1
    cc = pl*alpha + pc*(1-alpha);
elseif alpha<0
    cc = pc;
else
    cc = pl;
end
return

